from openstack.exceptions import ResourceNotFound

from interface.network.port_forwarding import PortForwardingInterface
from interface.network.resources import Port, FloatingIP, PortForwarding

from implementation.utils.connection import OpenStackConnection


class OSPortForwardingManager(PortForwardingInterface, OpenStackConnection):

    # TODO: gestionar els rangs de ports lliures/ocupats i floating ips? aquí o al django o on?
    # el delete el podem fer a partir de la ip i no de la instance_id


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.floating_ips = [
            FloatingIP(item) for item in self.openstack.list_floating_ips()]

        for floating_ip in self.floating_ips:
            items = self.openstack.network.floating_ip_port_forwardings(floating_ip.id)
            floating_ip.port_forwardings = [
                PortForwarding(item, floating_ip.address) for item in items]

        #self.port_forwardings = [
        #    PortForwarding(item) for item in self.openstack.network.port_forwardings(self.floating_ips[0].id)]


    def create_port_forwarding(self, instance_id, internal_port_number, protocol):
        """DUBTE hem de gestionar condicions com ara que no estigui creat ja
        un port forwarding igual, que la floating_ip no estigui activa, que no
        estigui associat a un router, etc.?
        """

        project_id = self.openstack.auth['project_id']

        # fetch internal Neutron port id of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # recorrer cada floating ip i cridar una funció que ens torni un port lliure o None, sinó crear nova ip
        floating_ip_id = self.floating_ips[0].id
        external_port = "8000" #get_free_port() from list of ports
        # floating_ip_id = "3179d648-9bb0-4871-854c-efffceb74fb0" #get_floating_ip_id

        self.openstack.network.create_port_forwarding(
            floatingip_id=floating_ip_id,
            internal_port_id=internal_port_id,
            internal_ip_address=internal_ip_address,
            internal_port=internal_port_number,
            external_port=external_port,
            protocol=protocol,
            description='')

        port_forwardings = [
            PortForwarding(item) for item in self.openstack.network.port_forwardings(
                self.floating_ips[0].id)]

        # DUBTE: l'instància de la connection sempre s'ha d'inicialitzar amb l'objecte auth.
        # no es podria compartir? Perquè aquí necessito accedir a funcions del port manager
        # (internal_port_id no coincideix amb la instance_id)
        # sinó s'haurà de fer a un nivell superior? de cercar-lo i passar-lo per paràmetre aquí
        # on ha d'estar aquesta lògica doncs?

        payload = {
            "port_forwarding": {
              	"project_id": f"{project_id}",
                "protocol": f"{protocol}",
                "internal_ip_address": f"{instance_id}",
                "internal_port": f"{internal_port_number}",
                "internal_port_id": f"{internal_port_id}",
                "external_port": f"{external_port}"
              }
        }

        # create a ks client and get token or do a post directly with client
        #ks = self.obtain_keystone_client()
        #print(ks)

        #r = requests.post("http://vm.niif.cloud.bme.hu:18686/v2.0/floatingips" \
        #    f"/{floating_ip_id}/port_forwardings", data = payload)
        #r.json()


    def delete_port_forwarding(self, instance_id, internal_port_number):
        """ Removes port forwarding rule
        """

        # fetch internal Neutron port of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # find floating ip where port forwarding rule is defined
        floating_ip_id = None
        for floating_ip in self.openstack.list_floating_ips():
            for port_forwarding in floating_ip.port_forwardings:
                if (port_forwarding.internal_ip_address == internal_ip_address)
                    and (port_forwarding.internal_port == internal_port_number):
                    # found
                    floating_ip_id = floating_ip.id
                    break
                    break

        # loop again to fetch port forwarding id
        # because is not contained in objects from the above results
        port_forwarding_id = None
        for port_forwarding in self.openstack.network.floating_ip_port_forwardings(floating_ip_id)
            if (port_forwarding.internal_ip_address == internal_ip_address)
                and (port_forwarding.internal_port == internal_port_number):
                port_forwarding_id = port_forwarding.id

        self.openstack.network.delete_port_forwarding(
            port_forwarding_id, floating_ip_id)


    def list_instance_port_forwardings(self, instance_id):
        """ Lists all port forwarding rules defined from a vm instance
        """

        # fetch internal Neutron port id of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # fetch port forwardings
        instance_port_forwardings = []
        for floating_ip in self.floating_ips:
            for item in self.openstack.network.port_forwardings(floating_ip.id):
                if item.internal_ip_address == internal_ip_address:
                    instance_port_forwardings.append(PortForwarding(item, floating_ip.address))

        return instance_port_forwardings


    def __get_internal_port_object_from_instance_id(self, instance_id):
        """ Fetches Neutron internal port object from instance_id
        """

        # query ports filtered by keyboard arguments (kwargs), returns a list
        results = [Port(item) for item in self.openstack.network.ports(
            device_id=instance_id)]

        # pick the result, we assume the list lenght is 1
        internal_port_object = results.pop()

        return internal_port_object
